/*
 * svn_date.c:  (natural language) date parsing routine.
 *              accepts most of the formats accepted by 
 *                 http://www.eyrx.com/DOC/D_GNUTIL/GSU_3.HTM
 *
 * ====================================================================
 * Copyright (c) 2000-2001 C. Scott Ananian.  All rights reserved.
 *
 * This software is licensed as described in the file COPYING, which
 * you should have received as part of this distribution.  The terms
 * are also available at http://subversion.tigris.org/license-1.html.
 * If newer versions of this license are posted there, you may use a
 * newer version instead, at your option.
 * ====================================================================
 */


#include <assert.h>
#include <stdlib.h>             /* for strtol */
#include <libintl.h>            /* for gettext */
#define APR_WANT_STRFUNC /* for strncmp, strlen, etc */
#define APR_WANT_MEMFUNC /* for memset, etc */
#define APR_WANT_STDIO   /* for sscanf */
#include <apr_want.h>    /* pulls in all the above */
#include <apr_lib.h>		/* apr_isdigit, apr_isspace */
#include <apr_time.h>           /* for apr_exploded_time_t & friends */
#include <apr_errno.h>          /* for apr_status_t, APR_SUCCESS, etc */
#include "svn_time.h"           /* for svn_parse_date prototype */


/*** Prototypes. */
/* parsers for possible date formats */
static apr_status_t
ISO8601ext_parse (const char *text, const apr_time_t now, apr_time_t *resultp);
static apr_status_t
ISO8601bas_parse (const char *text, const apr_time_t now, apr_time_t *resultp);
static apr_status_t
locale_parse (const char *text, const apr_time_t now, apr_time_t *resultp);

/* helper functions for the parsers */
static apr_status_t
ISO8601ext_parse_TZD (const char *text, apr_exploded_time_t *tm);
static apr_status_t
ISO8601bas_parse_TZD (const char *text, apr_exploded_time_t *tm);
static apr_status_t
locale_parse_time (const char *text, char **cp, apr_exploded_time_t *tm);
static apr_status_t
locale_parse_date (const char *text, char **cp, apr_exploded_time_t *tm);
static apr_status_t
locale_parse_day (const char *text, char **cp, apr_exploded_time_t *tm);
static apr_status_t
locale_parse_zone (const char *text, char **cp, apr_exploded_time_t *tm);
static apr_status_t
locale_parse_relative (const char *text, char **cp, apr_exploded_time_t *tm);

/* an enumerated type to select between local time and GMT */
typedef enum { SVN_LOCAL_TIME, SVN_GMT_TIME } timetype_t;


/* The one public interface of the date parser:  convert human-readable
   date TEXT into an apr_time_t.  Note that 'now' is passed as
   a parameter so that you can use this routine to find out how SVN
   *would have* parsed some string at some *arbitrary* time: relative
   times should always parse the same even if svn_parse_date is called
   multiple times during a computation of finite length.  For this reason,
   the 'now' parameter is *mandatory*. Returns APR_SUCCESS on success. */
apr_status_t
svn_parse_date (const char *text, const apr_time_t now, apr_time_t * result)
{
  apr_status_t status;
  /* this is a succession of candidate parsers.  the first one that
   * produces a valid parse of the *complete* string wins. */
  status = ISO8601ext_parse (text, now, result);
  if (status==APR_SUCCESS) goto done;
  status = ISO8601bas_parse (text, now, result);
  if (status==APR_SUCCESS) goto done;
  status = locale_parse (text, now, result);
  if (status==APR_SUCCESS) goto done;
  /* error: invalid date */
  status = APR_EBADDATE;
 done:
  return status;
}


/* Initialize 'apr_exploded_time_t' to all zeros (except for tm_mday field,
 * for which smallest valid value is 1).  */
static void
init_tm (apr_exploded_time_t *tm)
{
  memset (tm, 0, sizeof (*tm));
  tm->tm_mday = 1;              /* tm is not valid with all zeros. */
}

/* Compute the number of days since 1jan1970; the 'year' parameter is
 * "years since 1900".  Month must be in bounds! */
static apr_status_t
days_since_1jan1970(int *days_ptr, int year, int month, int mday) {
  static const int dayoffset[12] =
  {306, 337, 0, 31, 61, 92, 122, 153, 184, 214, 245, 275};
  int days;

  if (year < 70 || month < 0 || month > 11)
    return APR_EBADDATE;
  /* (borrowed from apr/time/time.c apr_implode_time() ) */

  /* shift new year to 1st March in order to make leap year calc easy */
  if (month < 2)
    year--;

  /* Find number of days since 1st March 1900 (in the Gregorian calendar). */
  days = year * 365 + year / 4 - year / 100 + (year / 100 + 3) / 4;
  days += dayoffset[month] + mday - 1;
  days -= 25508;              /* 1 jan 1970 is 25508 days since 1 mar 1900 */

  if (days < 0) return APR_EBADDATE;
  *days_ptr = days;
  return APR_SUCCESS;
}

/* we've got the following units:
 *  years, months, days, hours, minutes, seconds, microseconds.
 * There are always: 12 mon/year, 24 hr/day, 60 min/hr, 10^6 usec/sec.
 * There are variable numbers of days/month (leap years, etc) and
 * variable numbers of seconds/minute (leap seconds).
 * So we normalize the ones we can, and do a binary-search or ask an
 * oracle for the ones we can't. */
/* result is in GMT */
static apr_status_t
normalize_time(apr_exploded_time_t *tm, timetype_t local_or_gmt) {
  apr_exploded_time_t tmcopy = *tm, tmtest;
  apr_int64_t years, days, hours;
  apr_time_t lo, hi, x;
  apr_status_t status;
  int i;
  /* normalize all components in order from coarsest to finest,
   * ignoring yday and wday for now. */

  /* normalize months */
  years = tmcopy.tm_mon/12;
  if (tmcopy.tm_mon < 0)
    years--;
  tmcopy.tm_year += years;
  tmcopy.tm_mon -= 12*years;
  assert(tmcopy.tm_mon >= 0 && tmcopy.tm_mon < 12);
  /* we do the next normalizations cleverly to avoid overflow problems */
  /* normalize minutes */
  hours = tmcopy.tm_min / 60;
  if (tmcopy.tm_min < 0)
    hours--;
  tmcopy.tm_min += /*any extra minutes would go here*/- hours*60;
  assert(tmcopy.tm_min >= 0 && tmcopy.tm_min < 60);
  /* normalize hours */
  days = (hours+tmcopy.tm_hour)/24;
  if ((hours+tmcopy.tm_hour) < 0)
    days--;
  tmcopy.tm_hour += hours - days*24;
  assert(tmcopy.tm_hour >= 0 && tmcopy.tm_hour < 24);
  /* okay, now we have some extra days in 'days' */
  /* binary-search for a dates with the proper number of days. */

  /* desired days since 1jan1970: */
  if (APR_SUCCESS != 
      (status = days_since_1jan1970(&i, tmcopy.tm_year, tmcopy.tm_mon, 1)))
    goto fail;
  days += i + (tmcopy.tm_mday-1);
  /* lowest possible time_t (negative leap second every minute) */
  lo = days * (24*60*59); /* note: seconds, not usec */
  /* highest possible time_t (positive leap second every minute) */
  hi = days * (24*60*61); /* note: seconds, not usec */
  hi++; /* in routine below, hi actually contains 1 more than max poss. val */
  /* do the search */
  while (hi >= lo)
    {
      int found = 0;
      x = (hi+lo)/2;
      if (APR_SUCCESS != (status = apr_explode_gmt
                          (&tmtest, x * APR_USEC_PER_SEC)) ||
          APR_SUCCESS != (status = days_since_1jan1970
                          (&i, tmtest.tm_year, tmtest.tm_mon, tmtest.tm_mday)))
        goto fail;
      /* compare candidate tmtest with tmcopy */
      if (i < days ||
          (i==days &&
           (tmtest.tm_hour < tmcopy.tm_hour ||
            (tmtest.tm_hour==tmcopy.tm_hour &&
             (tmtest.tm_min < tmcopy.tm_min ||
              (tmtest.tm_min==tmcopy.tm_min &&
               (tmtest.tm_sec < 0 ||
                (tmtest.tm_sec==0 &&
                 !(found=1/*perfect match!*/)))))))))
        lo = x+1;
      else
        hi = x-1;
      if (found)
        {
          /* found match, so mday is now normalized.  Just account for the
           * time zone offset and the left-over seconds and microseconds. */
          x = (x+tmcopy.tm_sec-tmcopy.tm_gmtoff)*APR_USEC_PER_SEC
            + tmcopy.tm_usec;
          if (SVN_LOCAL_TIME == local_or_gmt)
            return apr_explode_localtime(tm, x);
          else
            return apr_explode_gmt(tm, x);
            
        }
    }
  /* match not found, something went wrong. */
  status = APR_EBADDATE;
 fail:
  return status;
}



/* Parse according to W3C subset of ISO 8601: see
 * http://www.w3.org/TR/NOTE-datetime
 * The complete date must be specified.
 * Any unspecified time components are assumed to be zeros.
 * This is the 'extended' format of ISO8601.
 * As per W3C subset, we don't allow week and day numbers. */
static apr_status_t
ISO8601ext_parse (const char *text, const apr_time_t now, apr_time_t * resultp)
{
  apr_exploded_time_t tm;
  char *cp;
  apr_time_t result;
  apr_status_t status;
  /* GENERAL EXTENSIONS TO W3C SUBSET: leading zeros are optional in
   * year, month, day, hour, minute, and second fields. We also
   * allow the locale's alternative numeric symbols, if any. */

  /* Try: YYYY-MM-DDThh:mm:ssTZD YYYY-MM-DDThh:mm:ss.sTZD */
  /* EXTENSION TO W3C SUBSET: Allow commas as decimal points as per ISO. */
  /* Allow YYYY-MM-DDThh:mm:ss,sTZD too */
  init_tm (&tm);
  if (APR_SUCCESS == apr_strptime (text, &cp, " %EY-%Om-%OdT%OH:%OM:%OS", &tm))
    {
      if (('.' == cp[0] || ',' == cp[0]) && apr_isdigit (cp[1]))
        {
          /* Discard fractional seconds. */
          for (cp++; apr_isdigit (*cp); cp++)
            ;
        }
      if (APR_SUCCESS == ISO8601ext_parse_TZD (cp, &tm))
        goto success;
    }

  /* Try: YYYY-MM-DDThh:mmTZD */
  init_tm (&tm);
  if (APR_SUCCESS == apr_strptime (text, &cp, " %EY-%Om-%OdT%OH:%OM", &tm) &&
      APR_SUCCESS == ISO8601ext_parse_TZD (cp, &tm))
    goto success;

  /* Try: YYYY-MM-DD */
  init_tm (&tm);
  if (APR_SUCCESS == apr_strptime (text, &cp, " %EY-%Om-%Od ", &tm)
      && 0 == *cp /* matched to end of string */)
    goto success;
  /* sketchy: we're parsing this as midnight UTC.  User may expect this
   * to be parsed as midnight localtime.  Should require that a time zone
   * designator be present, but this is not in ISO8601 spec. */

  /* The YYYY and YYYY-MM ISO8601 formats are disallowed as incomplete. */

  /* Give up: can't parse as ISO8601 extended format. */
  return APR_EBADDATE;

success:
  /* We succeeded in parsing as ISO8601 extended format! */
  status = apr_implode_gmt(&result, &tm);
  if (APR_SUCCESS == status)
    *resultp = result; /* success */
  return status;
}

/* Parse an ISO8601 extended Time Zone Designator from 'text'.
 * Update the 'apr_exploded_time_t' according to parsed time zone.  Returns
 * non-zero if we parse successfully and there is nothing but spaces
 * following the time zone designator. */
static apr_status_t
ISO8601ext_parse_TZD (const char *text, apr_exploded_time_t *tm)
{
  /* Time Zone Designator is 'Z' or '+hh:mm' or '-hh:mm' */
  apr_status_t status;
  int isplus = 0;
  switch (*(text++))
    {
    case 'Z':
      break;                    /* done! */
    case '+':
      isplus = 1;
    case '-':
      /* parse hh:mm */
      if (apr_isdigit (text[0]) && apr_isdigit (text[1]) && (text[2] == ':') &&
          apr_isdigit (text[3]) && apr_isdigit (text[4]))
        {
          /* we jump through hoops to use strtol here, in the hopes that
           * the c-library has localized this function (along with isdigit)
           * to deal with possibly non-roman digit sets */
          char buf[3] = { '0', '0', '\0' };
          int hours, minutes;
          buf[0] = text[0];
          buf[1] = text[1];
          hours = (int) strtol(buf, NULL, 10);
          buf[0] = text[3];
          buf[1] = text[4];
          minutes = (int) strtol(buf, NULL, 10);
          /* set gmtoff field (we'll renormalize to GMT at the end) */
          tm->tm_gmtoff = 60*(60*hours+minutes) * (isplus?1:-1);
          /* skip parsed hh:mm string */
          text += 5;            
          break;                /* done! */
        }
      /* EXTENSION TO W3C SUBSET: allow omitting the minutes section. */
      /* parse hh */
      if (apr_isdigit (text[0]) && apr_isdigit (text[1]))
        {
          /* we jump through hoops to use strtol here, in the hopes that
           * the c-library has localized this function (along with isdigit)
           * to deal with possibly non-roman digit sets */
          char buf[3] = { '0', '0', '\0' };
          int hours;
          buf[0] = text[0];
          buf[1] = text[1];
          hours = (int) strtol(buf, NULL, 10);
          /* set gmtoff field (we'll renormalize to GMT at the end) */
          tm->tm_gmtoff = 60*(60*hours+0) * (isplus?1:-1);
          /* skip parsed hh string */
          text += 2;
          break;                /* done! */
        }
    default:
      status = APR_EBADDATE;    /* fail */
      goto out;
    }
  /* renormalize to GMT */
  if (APR_SUCCESS != (status = normalize_time(tm, SVN_GMT_TIME)))
    goto out;
  /* should be only spaces until the end of string */
  while (apr_isspace (*text))
    text++;                     /* skip spaces */
  /* success if this is the end of string */
  status = (*text == '\0') ? APR_SUCCESS : APR_EBADDATE;
  /* done. */
 out:
  return status;
}


/* Parse according to 'basic' format of ISO8601.
 * We don't allow week and day numbers.
 */
static apr_status_t
ISO8601bas_parse (const char *text, const apr_time_t now, apr_time_t * resultp)
{
  apr_exploded_time_t tm;
  char *cp;
  apr_time_t result;
  apr_status_t status;
  /* extensions to ISO8601: leading zeros are optional in
   * year, month, day, hour, minute, and second fields. We also
   * allow the locale's alternative numeric symbols, if any. */

  /* Try: YYYYMMDDThhmmssTZD YYYYMMDDThhmmss.sTZD YYYYMMDDThhmmss,sTZD */
  init_tm (&tm);
  if (APR_SUCCESS == apr_strptime (text, &cp, " %EY%Om%OdT%OH%OM%OS", &tm))
    {
      if (('.' == cp[0] || ',' == cp[0]) && apr_isdigit (cp[1]))
        {
          /* Discard fractional seconds. */
          for (cp++; apr_isdigit (*cp); cp++)
            ;
        }
      if (APR_SUCCESS == ISO8601bas_parse_TZD (cp, &tm))
        goto success;
    }

  /* Try: YYYYMMDDThhmmTZD */
  init_tm (&tm);
  if (APR_SUCCESS == apr_strptime (text, &cp, " %EY%Om%OdT%OH%OM", &tm) &&
      APR_SUCCESS == ISO8601bas_parse_TZD (cp, &tm))
    goto success;

  /* Try: YYYYMMDD */
  init_tm (&tm);
  if (APR_SUCCESS == apr_strptime (text, &cp, " %EY%Om%Od ", &tm)
      && 0 == *cp /* matched to end of string */)
    goto success;
  /* sketchy: we're parsing this as midnight UTC.  User may expect this
   * to be parsed as midnight localtime.  Should require that a time zone
   * designator be present, but this is not in ISO8601 spec. */

  /* The YYYY and YYYYMM ISO8601 formats are disallowed as incomplete. */

  /* Give up: can't parse as ISO8601 basic format. */
  return APR_EBADDATE;

success:
  /* We succeeded in parsing as ISO8601 basic format! */
  status = apr_implode_gmt(&result, &tm);
  if (APR_SUCCESS == status)
    *resultp = result; /* success */
  return status;
}

/* Parse an ISO8601 basic Time Zone Designator from 'text'.
 * Update the 'apr_exploded_time_t' according to parsed time zone.  Returns
 * non-zero if we parse successfully and there is nothing but spaces
 * following the time zone designator. */
static apr_status_t
ISO8601bas_parse_TZD (const char *text, apr_exploded_time_t *tm)
{
/* Time Zone Designator is: 'Z' or '+hhmm' or '-hhmm' */
  apr_status_t status;
  int isplus = 0;
  switch (*(text++))
    {
    case 'Z':
      break;                    /* done! */
    case '+':
      isplus = 1;
    case '-':
      /* parse hhmm */
      if (apr_isdigit (text[0]) && apr_isdigit (text[1]) &&
          apr_isdigit (text[2]) && apr_isdigit (text[3]))
        {
          /* we jump through hoops to use strtol here, in the hopes that
           * the c-library has localized this function (along with isdigit)
           * to deal with possibly non-roman digit sets */
          char buf[3] = { '0', '0', '\0' };
          int hours, minutes;
          buf[0] = text[0];
          buf[1] = text[1];
          hours = (int) strtol(buf, NULL, 10);
          buf[0] = text[2];
          buf[1] = text[3];
          minutes = (int) strtol(buf, NULL, 10);
          /* set gmtoff field (we'll renormalize to GMT at the end) */
          tm->tm_gmtoff = 60*(60*hours+minutes) * (isplus?1:-1);
          /* skip parsed hhmm string */
          text += 4;
          break;                /* done! */
        }
      /* parse hh */
      if (apr_isdigit (text[0]) && apr_isdigit (text[1]))
        {
          /* we jump through hoops to use strtol here, in the hopes that
           * the c-library has localized this function (along with isdigit)
           * to deal with possibly non-roman digit sets */
          char buf[3] = { '0', '0', '\0' };
          int hours;
          buf[0] = text[0];
          buf[1] = text[1];
          hours = (int) strtol(buf, NULL, 10);
          /* set gmtoff field (we'll renormalize to GMT at the end) */
          tm->tm_gmtoff = 60*(60*hours+0) * (isplus?1:-1);
          /* skip parsed hh string */
          text += 2;
          break;                /* done! */
        }
    default:
      status = APR_EBADDATE;    /* fail */
      goto out;
    }
  /* renormalize to GMT */
  if (APR_SUCCESS != (status = normalize_time(tm, SVN_GMT_TIME)))
    goto out;
  /* should be only spaces until the end of string */
  while (apr_isspace (*text))
    text++;                     /* skip spaces */
  /* success if this is the end of string */
  status = (*text == '\0') ? APR_SUCCESS : APR_EBADDATE;
  /* done. */
 out:
  return status;
}


/* Parse 'human-readable' date, for the current locale.  Words not
 * numbers. =)  This does absolute and relative time/dates.  All times are
 * *LOCAL* unless time zone is specified.
 */
static apr_status_t
locale_parse (const char *text, const apr_time_t now, apr_time_t * resultp)
{
  apr_exploded_time_t tm;
  apr_status_t status;
  apr_time_t result;
  char *cp;

  /* initialize tm to 'now' in localtime */
  if (APR_SUCCESS != (status = apr_explode_localtime(&tm, now)))
    goto out;

  /* skip leading whitespace */
  while (apr_isspace (*text))
    text++;
  /* parse elements repeatedly until no elements are left (or we find
   * an unparsable string. */
  while (*text != '\0')
    {
      /* now attempt parse */
      if (APR_SUCCESS != locale_parse_time (text, &cp, &tm) &&
          APR_SUCCESS != locale_parse_date (text, &cp, &tm) &&
          APR_SUCCESS != locale_parse_day (text, &cp, &tm) &&
          APR_SUCCESS != locale_parse_zone (text, &cp, &tm) &&
          APR_SUCCESS != locale_parse_relative (text, &cp, &tm))
        /* can't parse as 'human-readable' absolute/relative date */
        return APR_EBADDATE;
      /* advance text pointer */
      text = cp;
      /* normalize tm to localtime */
      if (APR_SUCCESS != (status = normalize_time(&tm, SVN_LOCAL_TIME)))
        goto out;
      /* skip leading whitespace */
      while (apr_isspace (*text))
        text++;
    }

  if (APR_SUCCESS != (status = apr_implode_gmt(&result, &tm)))
    goto out;
  *resultp = result; /* success */
 out:
  return status;
}

/* Parse a time specification from the 'text' string.  Fill in fields
 * of 'tm' if successful and return a pointer to the end of the time
 * specification; else leave 'tm' untouched and return NULL. */
static apr_status_t
locale_parse_time (const char *text, char **retptr, apr_exploded_time_t *tm)
{
  static const char *formats[] = {
    /* note that formats are ordered from most-specific to least */
    "%OI:%OM:%OS %p ",
    "%OH:%OM:%OS ",
    "%OI:%OM %p ",
    "%OH:%OM ",
    "%OI %p ",
    /* Finally, try locale-specific time format. */
    "%EX ",
    NULL /* end of list */
  };
  apr_exploded_time_t tmcopy, tmtrial;
  char *cp;
  int i;
  /* Time fields are zero if not explicitly specified. */
  tmcopy = *tm;
  tmcopy.tm_usec = tmcopy.tm_sec = tmcopy.tm_min = tmcopy.tm_hour = 0;
  for (i=0; formats[i]!=NULL; i++)
    {
      tmtrial = tmcopy;
      if (APR_SUCCESS == apr_strptime (text, &cp, formats[i], &tmtrial))
        {
          /* success! */
          *tm = tmtrial;
          *retptr = cp;
          return APR_SUCCESS;
        }
    }
  /* failure */
  return APR_EBADDATE;
}

/* Parse a date specification from the 'text' string.  Fill in fields
 * of 'tm' if successful and return a pointer to the end of the date
 * specification; else leave 'tm' untouched and return NULL. */
static apr_status_t
locale_parse_date (const char *text, char **retptr, apr_exploded_time_t *tm)
{
  /* note that formats are ordered from most-specific to least */
  static const char *formats[] = {
    /* 1976-09-27: ISO 8601 */
    "%EY-%Om-%Od ",
    /* 27 September 1976 or 27sep1976, etc */
    "%Od %b %EY ",
    /* 27-September-1976 */
    "%Od-%b-%EY ",
    /* 27 sep */
    "%Od %b ",
    /* 27-sep */
    "%Od-%b ",
    /* September-27-1976 */
    "%b-%Od-%EY ",
    /* Sep 27, 1976 */
    "%b %Od, %EY ",
    /* Sep 27, */
    "%b %Od, ",
    /* Sep 27 */
    "%b %Od ",
    /* Sep-27 */
    "%b-%Od ",
    /* Finally, try locale-specific date format. */
    "%Ex ",
    NULL /* end of list */
  };
  apr_exploded_time_t tmtrial;
  char *cp;
  int i;
  /* All date fields default to current date unless specified. */
  for (i=0; formats[i]!=NULL; i++)
    {
      tmtrial = *tm;
      if (APR_SUCCESS == apr_strptime (text, &cp, formats[i], &tmtrial))
        {
          /* success! */
          *tm = tmtrial;
          *retptr = cp;
          return APR_SUCCESS;
        }
    }
  /* failure */
  return APR_EBADDATE;
}

/* Parse a day-of-week specification from the 'text' string.  Advance
 * date of 'tm' to that day-of-week if parse is successful and return
 * a pointer to the end of the day-of-week specification; else leave
 * 'tm' untouched and return NULL. */
static apr_status_t
locale_parse_day (const char *text, char **retptr, apr_exploded_time_t *tm)
{
  apr_exploded_time_t tmtrial = *tm;
  int daydiff = 0;
  char *cp;
  int i;
  /* Parse phrases like 'first monday', 'last monday', 'next monday'. */
  static const struct
  {
    const char *phrase;
    int weekdiff;
  } phrases[] = {
    /* These phrases should be translated: note that we call gettext below. */
    /* multiple adjective cases or variant forms can be separated with
     * semicolons. */
    { "last %a", -1 },
    { "%a; this %a", 0 },
    { "next %a; first %a", 1 },
    { "second %a", 2 },
    { "third %a", 3 },
    { "fourth %a", 4 },
    { "fifth %a", 5 },
    { "sixth %a", 6 },
    { "seventh %a", 7 },
    { "eighth %a", 8 },
    { "ninth %a", 9 },
    { 0, 0 }
  };
  /* Go through phrases and find the longest match. */
  cp = NULL;
  for (i = 0; phrases[i].phrase != NULL; i++)
    {
      char *tr_phrase = gettext (phrases[i].phrase);
      char buf[80];
      int j;
      while (*tr_phrase!='\0')
        {
          char *cptrial;
          /* separate out ';'-delimited phrases */
          for (j=0; *tr_phrase!='\0' && *tr_phrase!=';'; tr_phrase++)
            if (j < (sizeof(buf)-1))
              buf[j++]=*tr_phrase;
          buf[j++]='\0';
          /* skip delimiting semicolons */
          while (';' == *tr_phrase)
            tr_phrase++;
          /* try this phrase; ensure there is a space, punct. or
             end-of-string after the phrase; success only if
             this match ends up being longer than previous best. */
          if (APR_SUCCESS ==
              apr_strptime (text, &cptrial, buf, &tmtrial) &&
              (*cptrial=='\0' || *cptrial=='.' || *cptrial==',' ||
               apr_isspace(*cptrial)) &&
              (cp==NULL || cptrial > cp /* match must be longer */))
            {
              daydiff = 7 * phrases[i].weekdiff;
              cp = cptrial;
            }
          /* okay, go back up and try the next ;-delimited translated phrase */
        }
      /* okay, go back up and try next phrase list */
    }
  if (cp==NULL)
    /* failure: couldn't match any phrases */
    return APR_EBADDATE;
  /* else, success! */

  /* tmtrial.tm_wday has the day of the week.  Push current date ahead if
   * necessary. */
  if (tmtrial.tm_wday != tm->tm_wday) /* If selected day is not current day: */
    /* Compute the number of days we need to push ahead. */
    daydiff += (7 + tmtrial.tm_wday - tm->tm_wday) % 7;
  /* Increment day of the month. */
  tmtrial = *tm;
  tmtrial.tm_mday += daydiff;
  /* Renormalize tmtrial to localtime. */
  if (APR_SUCCESS != normalize_time(&tmtrial, SVN_LOCAL_TIME))
    return APR_EBADDATE;
  *tm = tmtrial;               /* This is the new date. */

  /* A period or comma following a day of the week item is ignored. */
  while (apr_isspace (*cp))
    cp++;
  if (*cp == ',' || *cp == '.')
    cp++;
  /* done! */
  *retptr = cp;
  return APR_SUCCESS;
}

/* Parse 'relative items' in date string, such as '1 year', '2 years
* ago', etc.  The strings in this function should be translated.
* Adjust 'tm' if parse is successful and return a pointer to the end
* of the 'relative item' specification; else leave 'tm' untouched and
* return NULL. */
static apr_status_t
locale_parse_relative (const char *text, char **retptr,
                       apr_exploded_time_t *tm)
{
  const char *cp;
  int i, secdiff, mindiff, mondiff;
  /* Phrases in this list should all be translated. */
  /* multiple adjective cases or variant forms can be separated with
   * semicolons. If some concept below is untranslatable, a zero-length
   * string may be used as the "translation". */
  const static struct
  {
    const char *phrase;
    int has_multiplier;
    /* A 'month' and a 'year' may be different numbers of seconds depending
     * on which month/year we're talking about; likewise a 'minute' may
     * include a leap second making it 59 or 61 seconds.  We express the
     * time using three canonical units: second, minutes, and months. */
    int seconds;
    int minutes;
    int months;
  }
  phrases[] = {
    /* semicolon-delimited phrases    has_mult  s  min          months */

    /* times in the past */
    { "%d second ago; %d seconds ago",       1,-1,   0,              0 },
    { "%d minute ago; %d minutes ago",       1, 0,  -1,              0 },
    { "%d hour ago; %d hours ago",           1, 0, -60,              0 },
    { "%d day ago; %d days ago",             1, 0, -60 * 24,         0 },
    { "%d week ago; %d weeks ago",           1, 0, -60 * 24 * 7,     0 },
    { "%d fortnight ago; %d fortnights ago", 1, 0, -60 * 24 * 7 * 2, 0 },
    { "%d month ago; %d months ago",         1, 0,   0,             -1 },
    { "%d year ago; %d years ago",           1, 0,   0,            -12 },
      /* times in the future */
    { "%d second; %d seconds",               1, 1,   0,              0 },
    { "%d minute; %d minutes",               1, 0,   1,              0 },
    { "%d hour; %d hours",                   1, 0,  60,              0 },
    { "%d day; %d days",                     1, 0,  60 * 24,         0 },
    { "%d week; %d weeks",                   1, 0,  60 * 24 * 7,     0 },
    { "%d fortnight; %d fortnights",         1, 0,  60 * 24 * 7 * 2, 0 },
    { "%d month; %d months",                 1, 0,   0,              1 },
    { "%d year; %d years",                   1, 0,   0,             12 },
      /* other special times */
    { "tomorrow",                            0, 0,  60 * 24,         0 },
    { "next week",                           0, 0,  60 * 24 * 7,     0 },
    { "next fortnight",                      0, 0,  60 * 24 * 7 * 2, 0 },
    { "next month",                          0, 0,   0,              1 },
    { "next year",                           0, 0,   0,             12 },
    { "yesterday",                           0, 0, -60 * 24,         0 },
    { "last week",                           0, 0, -60 * 24 * 7,     0 },
    { "last fortnight",                      0, 0, -60 * 24 * 7 * 2, 0 },
    { "last month",                          0, 0,   0,             -1 },
    { "last year",                           0, 0,   0,            -12 },
      /* the current time */
    { "today; now",                          0, 0,   0,              0 },
      /* nonce words */
    { "at; from",                            0, 0,   0,              0 },
      /* end of list */
    { 0,                                     0, 0,   0,              0 }
  };
  /* Try matching the phrases in the list; find the longest.
   * Note the use of %n in the format string to find out how many letters
   * sscanf has grokked over in its match. */
  cp = NULL;
  for (i = 0; phrases[i].phrase != NULL; i++)
    {
      char *tr_phrase = gettext (phrases[i].phrase);
      char buf[80];
      while (*tr_phrase!='\0')
        {
          int j, multiplier, chars = 0;
          /* separate out ';'-delimited phrases */
          for (j=0; *tr_phrase!='\0' && *tr_phrase!=';'; tr_phrase++ )
            if (j < (sizeof(buf)-3))
              buf[j++]=*tr_phrase;
          /* add '%n' to the end of the format string */
          buf[j++]='%';
          buf[j++]='n';
          buf[j++]='\0';
          /* skip delimiting semicolons */
          while (';' == *tr_phrase)
            tr_phrase++;
          /* some phrases have multiplier patterns, some do not. */
          multiplier = 1;
          if (phrases[i].has_multiplier)
            sscanf (text, buf, &multiplier, &chars);
          else
            sscanf (text, buf, &chars);
          /* if sscanf errors out before matching the final %n
           * pattern, then 'chars' will still be its initial 0 value.
           * We also want to make sure that there is a space or
           * end-of-string delimiter after our candidate phrase, and
           * that it's longer than any other matches found so far. */
          if (chars != 0 &&
              (text[chars] == '\0' || apr_isspace (text[chars])) &&
              (cp==NULL || text + chars > cp /* match must be longer */))
            {
              /* Found matching phrase! */
              secdiff = multiplier * phrases[i].seconds;
              mindiff = multiplier * phrases[i].minutes;
              mondiff = multiplier * phrases[i].months;
              cp = text + chars;
            }
          /* okay, go back up and try the next ;-delimited translated phrase */
        }
      /* okay, go back up and try next phrase list */
    }
  if (cp==NULL)
    /* failure: couldn't match any phrases */
    return APR_EBADDATE;
  /* else, success! */

  /* adjust tm */
  tm->tm_sec += secdiff;
  tm->tm_min += mindiff;
  tm->tm_mon += mondiff;
  /* Renormalize tm to localtime. */
  if (APR_SUCCESS != normalize_time(tm, SVN_LOCAL_TIME))
    return APR_EBADDATE;
  /* done */
  *retptr = (char*) cp; /* must lose const here */
  return APR_SUCCESS;
}

/* Parse a time-zone specification from the 'text' string.  Adjust
 * time of 'tm' to that time zone if parse is successful and return a
 * pointer to the end of the time-zone specification; else leave 'tm'
 * untouched and return NULL. */
static apr_status_t
locale_parse_zone (const char *text, char **retptr, apr_exploded_time_t *tm)
{
  /* this is ugly: we accept '-0100', '+0100', '-01:00', and '+01:00'
   * as well as 'GMT' 'UTC' and 'UT' for UTC.  We also allow a single-letter
   * 'military' time zone designation, so long as it is followed by a space
   * or end-of-string.
   * No other time zone designations are allowed; too many ambiguities.
   */
  const char *cp;
  int i;
  int minutediff;
  const static struct
  {
    const char *name;
    int hourdiff;
  } zones[] = {
    { "GMT", 0 },
    { "UTC", 0 },
    { "UT", 0 },
    { "A", 1 },
    { "B", 2 },
    { "C", 3 },
    { "D", 4 },
    { "E", 5 },
    { "F", 6 },
    { "G", 7 },
    { "H", 8 },
    { "I", 9 },
    { "K", 10 },
    { "L", 11 },
    { "M", 12 },
    { "N", -1 },
    { "O", -2 },
    { "P", -3 },
    { "Q", -4 },
    { "R", -5 },
    { "S", -6 },
    { "T", -7 },
    { "U", -8 },
    { "V", -9 },
    { "W", -10 },
    { "X", -11 },
    { "Y", -12 },
    { "Z", 0 },
    { 0, 0 }
  };
  /* Try time zone abbreviations from zones list. */
  for (i = 0; zones[i].name != NULL; i++)
    {
      int zonelen = strlen (zones[i].name);
      if (strncmp (text, zones[i].name, zonelen) == 0 &&
          (zonelen > 1 || apr_isspace(text[zonelen]) || text[zonelen] == '\0'))
        {
          cp = text + zonelen;
          minutediff = 60 * zones[i].hourdiff;
          goto success;
        }
    }
  /* Try +0000 +00:00 formats. */
  if ((text[0] == '-' || text[0] == '+') &&
      apr_isdigit (text[1]) && apr_isdigit (text[2]))
    {
      /* Again, jump through hoops to use strtol in order to accomodate
       * non-roman digit sets. */
      char buf[3] = { '0', '0', '\0' };
      buf[0] = text[1];
      buf[1] = text[2];
      minutediff = 60 * strtol (buf, NULL, 10);
      cp = text + 3;
      if (*cp == ':')
        cp++;
      if (apr_isdigit (cp[0]) && apr_isdigit (cp[1]))
        {
          buf[0] = cp[0];
          buf[1] = cp[1];
          minutediff += strtol (buf, NULL, 10);
          if (text[0] == '-')
            minutediff = -minutediff;
          cp += 2;
          goto success;
        }
    }
  /* failure */
  return APR_EBADDATE;

 success:
  tm->tm_gmtoff = minutediff * 60; /* seconds east of UTC */
  /* Done. */
  *retptr = (char*) cp; /* well known type hole: must lose const here */
  return APR_SUCCESS;
}


/* 
 * local variables:
 * eval: (load-file "../svn-dev.el")
 * end: */
